#include "plan9.h"
#include <draw.h>
#include <event.h>
#include "video.h"
#include "proto.h"
#include "dither.h"

enum
{
	Buflen = 80000,
};

int	quietFlag = 1;
int	qualityFlag = 0;
int	loopFlag = 0;
int	framerate = -1;
int	gammaCorrectFlag = 0;
double	gammaCorrect = 1.0;
int	chromaCorrectFlag = 0;
double	chromaCorrect = 1.0;
int	partialFlag;
int	startFrame = -1;
int	endFrame = -1;
int	audioFlag = 0;

void
main(int argc, char **argv)
{
	VidStream *vs;
	XInfo xi;
	char *p;
	int mark;

	ARGBEGIN{
	case 'a':
		audioFlag = 1;
		break;
	case 'f':
		p = ARGF();
		if(p == nil)
			goto Usage;
		framerate = atoi(p);
		break;
	case 'l':
		loopFlag = 1;
		break;
	case 'q':
		qualityFlag = 1;
		break;
	default:
    Usage:
		fprint(2, "usage: abaco [-f framerate] [-alq]\n");
		exits("usage");
	}ARGEND;

	lum_values = (int *) malloc(LUM_RANGE * sizeof(int) );
	cr_values = (int *) malloc(CR_RANGE * sizeof(int) );
	cb_values = (int *) malloc(CB_RANGE * sizeof(int) );

	init_tables();

/*	wpixel[0] = 0xf800;
	wpixel[1] = 0x07e0;
	wpixel[2] = 0x1f;
	xi.depth = 16;
*/
	wpixel[0] = 0xff0000;
	wpixel[1] = 0xff00;
	wpixel[2] = 0xff;
	xi.depth = 24;	


	InitColorDither(1);
	InitCrop();

	vs = NewVidStream(Buflen);
	vs->ppm_width = -1;
	vs->ppm_height = -1;
	vs->ppm_modulus = -1;
	vs->input = stdin;
	vs->seekValue = 0;
	vs->filename = "stdin";
	vs->ditherType = 0;
	vs->matched_depth = xi.depth;

	if(mpegVidRsrc(0, vs, 1, &xi) == 0)
		sysfatal("bad movie");

	vs->realTimeStart = ReadSysClock();
	for(;;){
		mark = vs->totNumFrames;
		while(mark == vs->totNumFrames)
			mpegVidRsrc(0, vs, 0, &xi);

		if(vs->film_has_ended){
			if(loopFlag){
				clear_data_stream(vs);
				ResetVidStream(vs);
				rewind(vs->input);
				if (vs->seekValue < 0)
					vs->seekValue = 0 - vs->seekValue;
			}
		}
		if(vs->film_has_ended)
			exits(nil);
	}
}

void
DoDitherImage(VidStream *vs)
{
	int h, w;
	uchar *l, *cr, *cb, *dp;

	l = vs->current->luminance,
	cr = vs->current->Cr,
	cb = vs->current->Cb,
	dp = vs->current->display;
	h = (int)vs->mb_height * 16;
	w = (int)vs->mb_width*16;

	Color32DitherImage(l, cr, cb, dp, h, w);
//	Color16DitherImage(l, cr, cb, dp, h, w);
}
