#include "plan9.h"
#include <draw.h>
#include <event.h>
#include "video.h"
#include "proto.h"
#include "dither.h"


unsigned char	pixel[256];
unsigned long	wpixel[256];

int	*lum_values;
int	*cr_values;
int	*cb_values;

/* Range values for lum, cr, cb. */
int	LUM_RANGE;
int	CR_RANGE;
int	CB_RANGE;

static int VidRateNum[16]={30, 24, 24, 25, 30, 30, 50, 60, 
                         60, 15, 30, 30, 30, 30, 30, 30};

void
resize(uint w, uint h)
{
	int fd;

	fd = open("/dev/wctl", OWRITE);
	if(fd >= 0){
		fprint(fd, "resize -dx %d -dy %d", w, h);
		close(fd);
	}
}

extern int framerate;

void
ExecuteDisplay(VidStream *vid_stream, XInfo *xinfo)
{
	static int	rate_deal = -1;
	static int	one_frame_time;
	static struct timeval tftarget, tfnow;
	register int	usec, sec;

	vid_stream->totNumFrames++;

	if(partialFlag){
		if(endFrame!=-1 && vid_stream->totNumFrames>endFrame){
			PrintTimeInfo(vid_stream);
			vid_stream->film_has_ended = TRUE;
			if(loopFlag)
				clear_data_stream(vid_stream);
			else 
				DestroyVidStream(vid_stream, xinfo);
			return;
		}
		if(vid_stream->totNumFrames < startFrame)
			return;
	}

	switch(rate_deal){
	case 0:    
		break;
	default:    
		gettimeofday(&tfnow, (struct timezone *)NULL);
		usec = tftarget.tv_usec - tfnow.tv_usec;
		sec  = tftarget.tv_sec - tfnow.tv_sec;
		if(usec < 0){
			usec += 1000000;
			sec--;
		}

		/* If we're not behind, wait a bit */
		if((sec >= 0)  &&  usec > 0){
			tfnow.tv_sec = sec;
			tfnow.tv_usec = usec;
			gettimeofday(&tfnow, (struct timezone *)NULL);
		}
		/* Setup target for next frame */
		tftarget.tv_usec = tfnow.tv_usec + one_frame_time;
		if(tftarget.tv_usec >= 1000000){
			tftarget.tv_usec -= 1000000;
			tftarget.tv_sec = tfnow.tv_sec + 1;
		}else 
			tftarget.tv_sec = tfnow.tv_sec;
		break;
	case -1:    
		switch(framerate){
		case -1:/* Go with stream Value */
			rate_deal = VidRateNum[vid_stream->picture_rate];
			gettimeofday(&tftarget, (struct timezone *)NULL);
			one_frame_time = 1000000 / rate_deal;
			break;
		case 0:/* as fast as possible */
			rate_deal = 0;
			break;
		default:      
			rate_deal = framerate;
			gettimeofday(&tftarget, (struct timezone *)NULL);
			one_frame_time = 1000000 / rate_deal;
			break;
		}
		break;
	}
	plan9play(vid_stream);
}

int
gettimeofday(struct timeval *tv)
{
	vlong n;

	n = nsec();
	tv->tv_sec = n/1000000000;
	tv->tv_usec = (n%1000000000)/1000;
	return 0;
}
